# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the basic IDMapper."""

class IDMapper:

    """Basic ID mapper."""

    def __init__(self, database):
        self.database = database
        self.objects = {}

    def get(self, model, primary):
        """
        Get an object from the ID mapper, or None.

        Args:
            model (Model): the model class.
            primary (dict): the primary fields.

        Returns:
            model (Model instance or None).

        """
        fields = tuple(primary.values())
        return self.objects.get(model, {}).get(fields)

    def set(self, model, primary, instance):
        """
        Set the object in the ID mapper.

        Args:
            model (Model): the model class.
            primary (dict): the primary fields.
            instance (Model): the model instance.

        """
        if self.get(model, primary):
            return

        fields = tuple(primary.values())
        objects = self.objects.get(model)
        if objects is None:
            objects = {}
            self.objects[model] = objects
        objects[fields] = instance

    def delete(self, model, primary):
        """
        Delete the specified model instance from the ID mapper.

        Args:
            model (Model): the model subclass.
            primary (dict): the primary field dictionary.

        """
        fields = tuple(primary.values())
        objects = self.objects.get(model)
        if objects is None:
            return

        return objects.pop(fields)



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the Transaction class, to handle database transactions."""

class Transaction:

    """
    Class representing an inner or outer transaction.
    """

    def __init__(self, database, parent=None):
        self.database = database
        self.engine = database._engine
        self.parent = parent
        self.objects = {}

    def __enter__(self):
        self.database._current_transaction = self
        self.engine.begin_transaction(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.database._current_transaction = self.parent
        if exc_type:
            # Restore transaction objects as they were.
            for obj, attrs in self.objects.items():
                obj._has_init = False
                for key, value in attrs.items():
                    setattr(obj, key, value)
                obj._has_init = True
            self.engine.rollback_transaction(self)
        else:
            self.engine.commit_transaction(self)



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test transactions."""

from test.base import BaseTest

from pygasus import Field, Model
from pygasus.exceptions import *

class Product(Model):

    """A simple product."""

    name = Field(str)
    price = Field(int)

class TestTransactions(BaseTest):

    """Test the transaction API."""

    def setUp(self):
        super().setUp()
        self.db.bind((Product, ))

    def test_create(self):
        """Create several instances."""
        # Try a transaction thqat should not fail.
        product_id = None
        with self.db.transaction:
            product = Product.create(name="apple", price=2)
            product_id = product.id

        # Right now, it should have been committed.
        self.assertIsNotNone(Product.get(id=product_id))

        # However, now try to add a product and cause a failure.
        try:
            with self.db.transaction:
                product = Product.create(name="pear", price=2)
                product_id = product.id
                raise InterruptedError
        except InterruptedError:
            pass

        # There should have been a rollback at this point, so
        # the product shouldn't be available.
        self.assertIsNone(Product.get(id=product_id))

    def test_update(self):
        """Test to update within a transaction."""
        product_id = None
        product = Product.create(name="apple", price=2)
        product_id = product.id

        # Try to change the price.
        with self.db.transaction:
            product.price = 3

        # Check that the product has been updated.
        product = Product.get(id=product_id)
        self.assertEqual(product.price, 3)

        # Now try a new update, but make sure it fails.
        try:
            with self.db.transaction:
                product.price = 4
                raise InterruptedError
        except InterruptedError:
            pass

        # There should have been a rollback, so the price
        # shouldn't have changed.
        product = Product.get(id=product_id)
        self.assertEqual(product.price, 3)

    def test_inner(self):
        """Test an inner transaction."""
        first_id = second_id = None
        with self.db.transaction:
            product = Product.create(name="apple", price=2)
            first_id = product.id

            # Open an inner transaction.
            with self.db.transaction:
                product = Product.create(name="apple", price=2)
                second_id = product.id

        # Since there has been no exception, both products should exist.
        self.assertIsNotNone(Product.get(id=first_id))
        self.assertIsNotNone(Product.get(id=second_id))

        # Now do the same thing, raising an exception in the inner block.
        with self.db.transaction:
            product = Product.create(name="apple", price=2)
            first_id = product.id

            # Open an inner transaction.
            try:
                with self.db.transaction:
                    product = Product.create(name="apple", price=2)
                    second_id = product.id
                    raise InterruptedError
            except InterruptedError:
                pass

        # The first product should exist, the second one should not.
        self.assertIsNotNone(Product.get(id=first_id))
        self.assertIsNone(Product.get(id=second_id))

        # Do the same with an error in the outer block.
        try:
            with self.db.transaction:
                product = Product.create(name="apple", price=2)
                first_id = product.id

                with self.db.transaction:
                    product = Product.create(name="apple", price=2)
                    second_id = product.id
                raise InterruptedError
        except InterruptedError:
            pass

        # No product should exist.
        self.assertIsNone(Product.get(id=first_id))
        self.assertIsNone(Product.get(id=second_id))



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test the ID mapper."""

from test.base import BaseTest

from pygasus import IDMapper, Field, Model
from pygasus.exceptions import *

class Car(Model):

    """A simple car."""

    name = Field(str)
    price = Field(int)

class TestIDMapper(BaseTest):

    """Test the transaction API."""

    def setUp(self):
        super().setUp()
        self.db.id_mapper = IDMapper(self)
        self.db.bind((Car, ))

    def test_create_and_get(self):
        """Create several cars, checking their identities."""
        ford = Car.create(name="Ford", price=10000)
        peugeot = Car.create(name="Peugeot", price=8000)
        pygasus = Car.create(name="Pygasus", price=15000)
        self.assertIs(ford, Car.get(id=ford.id))
        self.assertIs(peugeot, Car.get(id=peugeot.id))
        self.assertIs(pygasus, Car.get(id=pygasus.id))

    def test_create_in_transaction(self):
        """Create and get in transactions."""
        # Create within a first transaction, no error.
        with self.db.transaction:
            ford = Car.create(name="Ford", price=10000)
            peugeot = Car.create(name="Peugeot", price=8000)
            pygasus = Car.create(name="Pygasus", price=15000)

        # No error has occurred, so the result should be the same as above.
        self.assertIs(ford, Car.get(id=ford.id))
        self.assertIs(peugeot, Car.get(id=peugeot.id))
        self.assertIs(pygasus, Car.get(id=pygasus.id))

        # Create an inner transaction with an error.
        with self.db.transaction:
            ford = Car.create(name="Ford", price=10000)
            peugeot = Car.create(name="Peugeot", price=8000)

            try:
                with self.db.transaction:
                    pygasus = Car.create(name="Pygasus", price=15000)
                    raise InterruptedError
            except InterruptedError:
                pass

        # An error has occurred in the inner transaction, so the first
        # and second car should be fine, but there shouldn't be any third car.
        self.assertIs(ford, Car.get(id=ford.id))
        self.assertIs(peugeot, Car.get(id=peugeot.id))
        self.assertIsNone(Car.get(id=pygasus.id))

    def test_update(self):
        """Test to update fields from cars."""
        # Updating without a transaction should behave as expected.
        ford = Car.create(name="Ford", price=10000)
        peugeot = Car.create(name="Peugeot", price=8000)
        pygasus = Car.create(name="Pygasus", price=15000)
        pygasus.price = 12000

        # Check the price in instance (ID mapped).
        self.assertEqual(ford.price, 10000)
        self.assertEqual(peugeot.price, 8000)
        self.assertEqual(pygasus.price, 12000)

        # Get the cars from the database.  The price should be the same.
        ford = Car.get(id=ford.id)
        peugeot = Car.get(id=peugeot.id)
        pygasus = Car.get(id=pygasus.id)
        self.assertEqual(ford.price, 10000)
        self.assertEqual(peugeot.price, 8000)
        self.assertEqual(pygasus.price, 12000)

        # Now, modify within a transaction without error.
        with self.db.transaction:
            pygasus.price = 14000
            self.assertEqual(pygasus.price, 14000)

        self.assertEqual(pygasus.price, 14000)
        self.assertEqual(Car.get(id=pygasus.id).price, 14000)

        # Make two modifications.  The one in the inner block should
        # be cancelled, because of an error.
        with self.db.transaction:
            pygasus.price = 17000
            self.assertEqual(pygasus.price, 17000)

            try:
                with self.db.transaction:
                    pygasus.price = 18000
                    self.assertEqual(pygasus.price, 18000)
                    raise InterruptedError
            except InterruptedError:
                pass

        self.assertEqual(pygasus.price, 17000)
        self.assertEqual(Car.get(id=pygasus.id).price, 17000)
